-- Vital 2.1
--
import SS
--
-- LOCN 825 75 true
-- STYLE 
{-
Defining the "Look and Feel" of an abstract type
-}
-- LOCN 50 75 true
-- STYLE title
--
import ColorRowLib
--
-- LOCN 75 425 true
-- STYLE 
{-
This document describes in detail how the "Edit" type class can be used to define
the "Look and Feel" of an abstract data type so that:

   -  values of that type are automatically displayed in pictorial form, and

   -  the end-user can edit values of that type by simple point-and-click mouse operations,
      with the result of these edits being permanently incorporated in the
      Haskell source code of the document.
-}
-- LOCN 50 150 true
-- STYLE commentB
{-
A function
    mkRow ::  Int  ->  Row
that creates a row of coloured squares. For example:
-}
-- LOCN 125 550 true
-- STYLE comment
--
row = mkRow 8
--
-- LOCN 125 625 true
-- STYLE Hide
{-
... and a function
      rowToPic
that yields a pictorial representation of a row. For example:
-}
-- LOCN 125 800 true
-- STYLE comment
{-
... and functions 
      setYellow,  setCyan  ::  Int  ->  Row  ->  Row
to set the colour of a given square. For example:
-}
-- LOCN 125 675 true
-- STYLE comment
--
row1 = setCyan 3 (setCyan 1 row)
--
-- LOCN 125 750 true
-- STYLE Hide
--
pic = rowToPic row1
--
-- LOCN 125 875 true
-- STYLE 
{-
By (explicitly) using the "Picture" style, such a value can be displayed as a picture:
-}
-- LOCN 50 975 true
-- STYLE comment
--
pic
--
-- LOCN 75 1025 true
-- STYLE Picture
{-
Using the "Edit" type class to define the "Look" of an ADT
-}
-- LOCN 50 1200 true
-- STYLE heading1
--
row2 = (setCyan 1 . setYellow 4 . setCyan 5 . setCyan 4 . setCyan 6)  (mkRow 10)
--
-- LOCN 75 1525 true
-- STYLE 
{-
(Click on the handle of an expression [ie, 
 the small square beside it] to see its style
 displayed in the "Style" field at the top of
 the window.)
-}
-- LOCN 675 1050 true
-- STYLE comment
{-
Using the "Edit" class to define the "Feel" of an ADT
-}
-- LOCN 50 1675 true
-- STYLE heading1
{-
The "Edit" class provides the means to define the "feel" of a type, that is, the way that values of
the type respond to mouse gestures.
-}
-- LOCN 50 1725 true
-- STYLE commentB
{-
As an example, we will provide the "Row" type with the means to allow the user to set the colour of any
of its squares and (purely for illustration) the means to inspect the "path" to any of its squares.
-}
-- LOCN 50 1800 true
-- STYLE comment
{-
There are two stages involved in defining the way values of a type respond to mouse gestures:

    -  defining the menu of commands that the user is offered;

    -  implementing the "editing functions" that carry out these commands.
-}
-- LOCN 50 1875 true
-- STYLE comment
{-
Defining the menu of commands for an ADT
-}
-- LOCN 50 2050 true
-- STYLE heading1
{-
The "edit" method of the "Edit" class is used to define the menu of commands that appears when a user
right-clicks on the term (or, more carefully stated, when a user right-clicks within any "selectable rectangle"
in the visual representation for the ADT).
-}
-- LOCN 50 2100 true
-- STYLE comment
{-
The "edit" method of the "Edit" class is of type:

    edit  ::  a  ->  Path  ->  [(String,  EditFn)]

It takes as arguments the value of the expression and the "path" to the selected subterm, and it yields 
a list of pairs:

  - the first element of each pair defines the command name that will appear in the drop-down menu;

  - the second element defines the editing function that will be invoked (by the system) to carry out the command.
-}
-- LOCN 50 2200 true
-- STYLE comment
--
instance Edit Row where 
  edit board path = 
    [("Set yellow", setYellowEditFn),
     ("Set cyan",   setCyanEditFn),
     ("Show path",  showPathEditFn)]

--
-- LOCN 75 2525 true
-- STYLE 
{-
Here is an instance declaration for the "Row"  ADT that specifies three commands:
  - Set a square to yellow,
  - Set a square to cyan,
  - Display the value of the "path" to the square
-}
-- LOCN 50 2425 true
-- STYLE comment
{-
Try this on the following value (of type Row):
  - double-click on one of the coloured squares to Select the element
  - then right-click to drop-down the menu.
-}
-- LOCN 50 2650 true
-- STYLE commentB
--
row2
--
-- LOCN 75 2750 true
-- STYLE 
{-
Defining the Editing Functions for the ADT
-}
-- LOCN 50 2900 true
-- STYLE heading1
{-
The menu defined above specifies that one these three Editing Functions:
    - setYellowEditFn
    - setCyanEditFn
    - showPathEditFn
will be invoked (depending on which menu entry the user selects).
-}
-- LOCN 50 2950 true
-- STYLE comment
{-
The role of an Editing Function is to update the Haskell source text of the program so that, when
evaluated, it gives rise to the appropriate value.  (Note very carefully that it does NOT directly
change the value itself -- only the source code that gives rise to the value.)
-}
-- LOCN 50 3075 true
-- STYLE commentB
{-
. . . and it takes 7 arguments:
  - the value of the overall expression;
  - the "path" to the sub-picture the user selected (by double-clicking);
  - the X and the Y coords of the mouse;
  - the LHS (if any) of the Haskell declaration;
  - the RHS of the Haskell declaration;
  - the contents (if any) of the clipboard

and it yields a triple of Maybe Strings:
  - an updated RHS for the Haskell declaration;
  - an updated value for the clipboard;
  - a message to display in the Status line at the bottom of the window.

Typically, only one or two arguments are actually used, and only one or two of the three
possible result strings are generated.
-}
-- LOCN 50 3300 true
-- STYLE comment
{-
Here is the definition of the "showPathEditFn" function. Its purpose is to
display (for tutorial purposes) the value of the "path" (a list of integers)
to the component the user has selected. Thus, it uses only the "path" argument
and it defines only the "message" component of its output:
-}
-- LOCN 50 3625 true
-- STYLE comment
--
showPathEditFn row path x y lhs rhs clip =
  let message' = show path
  in (Nothing, Nothing, Just message')
--
-- LOCN 75 3750 true
-- STYLE Hide
--
showPathEditFn :: EditFunction
--
-- LOCN 75 3725 true
-- STYLE 
{-
. . . and here are the definitions for the edit functions associated with the commands to
set the colour of a square to yellow or cyan:
-}
-- LOCN 50 3850 true
-- STYLE comment
--
setYellowEditFn row path x y lhs rhs clip =
  let {i = decode path;
       rhs' = "setYellow " ++ show i ++ " (" ++ rhs ++ ")";
       message' = "Square set to yellow"
  } in (Just rhs', Nothing, Just message')
--
-- LOCN 75 3925 true
-- STYLE Hide
--
setCyanEditFn row path x y lhs rhs clip =
  let {i = decode path;
       rhs' = "setCyan " ++ show i ++ " (" ++ rhs ++ ")";
       message' = "Square set to cyan"
  } in (Just rhs', Nothing, Just message')
--
-- LOCN 75 4050 true
-- STYLE Hide
{-
Finally, it remains only to define the "decode" function. The role of this
function is to convert the path (a list of integers) to the selected component
in the visual representation to the corresponding element position (0, 1, 2, etc) in the
row of squares. Inspection of the way the visual representation is formed reveals
that the decoding function in this case is simply:
-}
-- LOCN 50 4175 true
-- STYLE comment
--
decode :: Path -> Int
--
-- LOCN 75 4300 true
-- STYLE 
--
decode path = length path - 2
--
-- LOCN 75 4325 true
-- STYLE Hide
{-
Although this document is a long one, the majority of it is commentary; it has
not taken many lines of code to define either the Look or the Feel of the "Row" ADT.
The code for performing a wider range of operations on a more complex ADT is likely
to be significantly longer -- but it will follow identical principles to the code shown here.
-}
-- LOCN 50 4800 true
-- STYLE commentB
{-
Testing the Look and Feel of the ADT
-}
-- LOCN 50 4400 true
-- STYLE heading1
{-
Try out the editing commands on the following declaration, examining the way the
declaration is modified by each command:
-}
-- LOCN 50 4475 true
-- STYLE commentB
--
testRow = mkRow 12
--
-- LOCN 75 4550 true
-- STYLE 
{-
Back to previous document
-}
-- LOCN 50 4925 true
-- STYLE open ""
{-
Back to previous document
-}
-- LOCN 750 150 true
-- STYLE open ""
{-
As a running example, we will take an implementation (defined in the libary module "ColorRowLib")
of an abstract type, "Row", that represents the (near trivial!) notion of "a row of coloured squares".
-}
-- LOCN 50 375 true
-- STYLE comment
{-
In addition to the abstract type "Row", this module exports the following entities:
-}
-- LOCN 50 500 true
-- STYLE comment
{-
As explained in an accompanying document, values of type Row can automatically
be displayed in this pictorial form simply by making use of the "View" style 
constructor to define an appropriate instance of the "style" method of the Edit class, 
thus:
-}
-- LOCN 50 1275 true
-- STYLE comment
--
instance Edit Row where
  style row = View rowToPic Picture
--
-- LOCN 75 1375 true
-- STYLE 
{-
For example, the following value of the abstract type Row is automatically displayed in pictorial form:
-}
-- LOCN 50 1475 true
-- STYLE comment
{-
(Follow this link for explanation
 of the principles involved.)
-}
-- LOCN 725 1275 true
-- STYLE open "EditStyleDemo"
{-
An editing function is of type:
-}
-- LOCN 50 3175 true
-- STYLE comment
--
type EditFunction = Row -> Path -> Int -> Int -> String -> String -> String -> (MString, MString, MString)
--
-- LOCN 75 3225 true
-- STYLE 
--
type MString = Maybe String
--
-- LOCN 75 3250 true
-- STYLE 
{-
(Note that these functions update the _text_ 
 of a declaration - and hence (indirectly) 
 its value.)
-}
-- LOCN 600 4000 true
-- STYLE comment
{-
(The "path" to a selected component in a
 picture is a list of {0, 1} values 
 describing its location in the abstract
 syntax of the Pic value.)
-}
-- LOCN 650 3625 true
-- STYLE comment
{-
(Double-click to Select an element, then right-click to drop-down a menu. Use 
 the "Undo" button on the toolbar to return to an earlier state.)
-}
-- LOCN 250 4675 true
-- STYLE comment
{-
The library module "ColorRowLib"
-}
-- LOCN 750 425 true
-- STYLE openLib "ColorRowLib"
{-
(Use the "Back" button
 on the toolbar to return
 to this document.)
-}
-- LOCN 750 475 true
-- STYLE comment
{-
A detailed account of how the "look and feel" of an ADT can be
specified using the Edit class. Illustrated by an ADT consisting
of a sequence of colours, displayed as a row of coloured squares.
-}
-- LOCN 50 -75 true
-- STYLE abstract
