-- Vital 2.1
--
mkBoard n cs ss =
    let { row = cycle [Rect ss ss c black | c <- cs];
          board = iterate tail row
    } in (above . take n . map (beside . take n)) board
--
-- LOCN 50 2225 true
-- STYLE Hide
{-
Solving the "n-queens" problem
-}
-- LOCN 50 100 true
-- STYLE title
--
import SS
--
-- LOCN 700 50 true
-- STYLE 
--
import Pictures
--
-- LOCN 700 100 true
-- STYLE 
--
mkBoard 6 [black, blue, gray, green, red] 30
--
-- LOCN 50 2400 true
-- STYLE Picture
--
mkBoard 8 [yellow, brown] 20
--
-- LOCN 550 2400 true
-- STYLE Picture
{-
First we need a function that will draw an n x n board.  Here is a 
fairly general one:
-}
-- LOCN 50 2125 true
-- STYLE comment
{-
Arguments:
  - n: the number of squares
  - cs: a list of colours for the squares
  - ss: the size of a single square
-}
-- LOCN 625 2200 true
-- STYLE comment
--
mkBoard :: Int -> [Color] -> Int -> Pic
--
-- LOCN 50 2200 true
-- STYLE 
{-
For example:
-}
-- LOCN 50 2350 true
-- STYLE comment
{-
The n-queens problem is a generalisation of the ancient and famous "8 queens" problem. This latter
involves placing 8 queens on an chess board in such a way that no queen is threatened by any other.
-}
-- LOCN 50 200 true
-- STYLE comment
{-
Here is a function that takes n as a parameter and yields a list of all possible
solutions to the problem. It is defined in terms of a subsidiary function, queens'.
-}
-- LOCN 50 625 true
-- STYLE comment
{-
The "n queens" problem involves placing n queens on an n x n board in such a way that
no queen is threated by any other.

Since (obviously!) there cannot be more than one queen on any column of the
board, and there are n queens in total, there must be exactly one queen in each
column. Thus, any solution to this problem can be represented by a list of n
integers (each in the range 0 to (n-1)) representing the position of the
queen in each column.
-}
-- LOCN 50 275 true
-- STYLE comment
{-
It is useful to introduce some types to represent boards positions and solutions:
-}
-- LOCN 50 475 true
-- STYLE comment
--
type Col = Int
--
-- LOCN 50 525 true
-- STYLE 
--
type Position = [Col]
--
-- LOCN 50 550 true
-- STYLE 
--
type Solutions = [Position]
--
-- LOCN 50 575 true
-- STYLE 
--
queens :: Int -> Solutions
--
-- LOCN 50 700 true
-- STYLE 
--
queens n = queens' n n
--
-- LOCN 50 725 true
-- STYLE Hide
{-
The function queens' develops the solution recursively. The value of
    queen' n m

is the list of solutions to the problem of placing a queen in each of the 
first m columns of an  n x n board in such a way that no queen can capture any
other queen.
-}
-- LOCN 50 825 true
-- STYLE comment
--
queens' n m =
  let {cols = [0 .. (n-1)];
       q = \m -> if m == 0 then [[]]
                 else concat [[col:board | col <- cols, safe col board 1] | board <- q (m-1)]

  } in q m
--
-- LOCN 50 1025 true
-- STYLE Hide
--
queens' :: Int -> Int -> Solutions
--
-- LOCN 50 1000 true
-- STYLE 
{-
The subsidiary function, safe, determines whether or not all queens are
safe from capture. The value of
    safe x ys 1

is True iff a queen placed in row x of column 0 is safe from capture by queens
in succeeding columns in the positions described by ys.
-}
-- LOCN 50 1175 true
-- STYLE comment
--
safe x ys m =
  case ys of {
    []    -> True;
    y:ys'  -> (x /= y) && (x /= (y-m)) && (x /= (y+m)) && safe x ys' (m+1)
  }
--
-- LOCN 50 1350 true
-- STYLE Hide
--
safe :: Int -> [Int] -> Int -> Bool
--
-- LOCN 50 1325 true
-- STYLE 
--
queens 1
--
-- LOCN 50 1550 true
-- STYLE 
--
queens 2
--
-- LOCN 250 1550 true
-- STYLE 
--
queens 3
--
-- LOCN 425 1550 true
-- STYLE 
--
queens 4
--
-- LOCN 575 1550 true
-- STYLE 
{-
Here, for example, are the sets of solutions to this problem for boards of sizes 1, 2, 3 and 4.
-}
-- LOCN 50 1500 true
-- STYLE comment
{-
It shows that:
    - there is exactly one way (!) of placing a queen safely on a 1 x 1 board
    - there is no way of placing 2 queens safely on a 2 x 2 board
    - . . . and likewise with a 3 x 3 board
    - there are two ways of placing 4 queens safely on a 4 x 4 board
       (in fact, they are mirror images of each other)
-}
-- LOCN 50 1750 true
-- STYLE comment
{-
Displaying a solution to the n-queens problem
-}
-- LOCN 50 2025 true
-- STYLE heading1
{-
Exercise 5.1
-}
-- LOCN 50 1925 true
-- STYLE open2'  "PicEx"
{-
We now develop a way of displaying solutions to the n-queens problem
-}
-- LOCN 50 2075 true
-- STYLE comment
{-
Next, we need a function that will produce a picture of
a given position. For this, we can use a variant of the function
developed in an earlier module.
-}
-- LOCN 50 2700 true
-- STYLE comment
--
placeQueens :: Int -> Position -> Pic
--
-- LOCN 50 2800 true
-- STYLE 
--
placeQueens ss cols =
  let queen = mkQueen ss
  in case cols of {
       []         -> NoPic;
       col:cols'  -> let { pic1 = move 0 (ss*col) queen;
                           pic2 = placeQueens ss cols'
                     } in pic1 <+> move ss 0 pic2
     }
--
-- LOCN 50 2825 true
-- STYLE Hide
{-
For example, the placement of queens for the first solution of
the 4-queens problem (above) is:
-}
-- LOCN 50 3000 true
-- STYLE comment
--
placeQueens 20 soln
--
-- LOCN 50 3075 true
-- STYLE Picture
--
mkQueen :: Int -> Pic
--
-- LOCN 650 2875 true
-- STYLE 
--
mkQueen ss = 
  let sf = fromInt ss / fromInt 40
  in scale sf queen'
--
-- LOCN 650 2900 true
-- STYLE Hide
{-
The subsidiary  "mkQueen" function used
here yields a picture of a queen scaled
to fit the size of a square:
-}
-- LOCN 650 2800 true
-- STYLE comment
--
mkDiagram ss c1 c2 posn = 
    let { n    = length posn;
          pic1 = placeQueens ss posn;
          pic2 = mkBoard n [c1, c2] ss
    } in pic2 <+> pic1
--
-- LOCN 50 3325 true
-- STYLE Hide
--
soln = head (queens 4)
--
-- LOCN 575 1650 true
-- STYLE 
--
mkDiagram 20 yellow brown soln
--
-- LOCN 50 3500 true
-- STYLE Picture
--
mkDiagram :: Int -> Color -> Color -> Position -> Pic
--
-- LOCN 50 3300 true
-- STYLE 
{-
We can now combine the function that creates a picture of a board with
the one that layouts the queens and so obtain a function that creates
a diagram of a potential solution to an n-queens problem.
-}
-- LOCN 50 3225 true
-- STYLE comment
{-
For example:
-}
-- LOCN 50 3450 true
-- STYLE comment
--
mkDiagram 40 cyan blue soln
--
-- LOCN 525 3500 true
-- STYLE Picture
{-
Exercise 5.2
-}
-- LOCN 50 3750 true
-- STYLE open2' "PicEx"
{-
Displaying all solutions to an n-queens problem
-}
-- LOCN 50 3875 true
-- STYLE heading1
--
mkDiagramAll ss c1 c2 n =
    let { solutions = queens n;
          f = move ss 0 . mkDiagram ss c1 c2;
          diagrams = map f solutions
    } in beside diagrams
--
-- LOCN 50 4050 true
-- STYLE Hide
--
diagrams = mkDiagramAll 20 yellow green
--
-- LOCN 50 4350 true
-- STYLE Hide
{-
The  "queens" function (above) delivered the list of all solutions to the n-queens problem. The
following function displays these pictorially:
-}
-- LOCN 50 3950 true
-- STYLE comment
--
mkDiagramAll :: Int -> Color -> Color -> Int -> Pic
--
-- LOCN 50 4025 true
-- STYLE 
{-
Arguments:
  ss: the square size
  c1: colour of the "white" squares
  c2: colour of the "black" squares
  n:  size of the board
-}
-- LOCN 600 4025 true
-- STYLE comment
{-
The function can be specialised for a particular square size and
colours. For example:
-}
-- LOCN 50 4250 true
-- STYLE comment
--
diagrams :: Int -> Pic
--
-- LOCN 50 4325 true
-- STYLE 
{-
. . . and, using this function, solutions to the various instances (n = 0, 1, 2, ...)
of the n-queens problem can be displayed.
-}
-- LOCN 50 4425 true
-- STYLE comment
--
diagrams 0
--
-- LOCN 50 4575 true
-- STYLE Picture
--
diagrams 1
--
-- LOCN 275 4575 true
-- STYLE Picture
--
diagrams 2
--
-- LOCN 500 4575 true
-- STYLE Picture
--
diagrams 3
--
-- LOCN 700 4575 true
-- STYLE Picture
{-
Boards of size 0, 2 and 3 have no solutions; a board of size 1 has exactly one solution:
-}
-- LOCN 50 4525 true
-- STYLE comment
{-
A board of size 4 has 2 solutions (which are, of course, symmetrically related):
-}
-- LOCN 50 4700 true
-- STYLE comment
--
diagrams 4
--
-- LOCN 50 4750 true
-- STYLE Picture
{-
A board of size 5 has 10 solutions (you will need to scroll sideways to see all of them):
-}
-- LOCN 50 4925 true
-- STYLE comment
--
diagrams 5
--
-- LOCN 50 4975 true
-- STYLE Picture
{-
. . . and there are some 92 solutions (including many symmetrically related) for
the standard 8 x 8 chessboard:
-}
-- LOCN 50 5175 true
-- STYLE comment
--
diagrams 8
--
-- LOCN 50 5250 true
-- STYLE Picture
{-
Exercise 5.3
-}
-- LOCN 50 5500 true
-- STYLE open2 "PicEx_5_3"
{-
Back to previous document
-}
-- LOCN 75 5625 true
-- STYLE open1 ""
--
import SSExercises
--
-- LOCN 700 75 true
-- STYLE 
{-
Back to previous document
-}
-- LOCN 700 150 true
-- STYLE open1 ""
{-
Incrementally builds up a function that graphicallydisplays solutions to
the n-queens problem.
-}
-- LOCN 50 -50 true
-- STYLE abstract
