-- Vital 2.1
{-
A case study, illustrating the nature of DNA and the Genetic Code, using
definitions from the DNA_lib module.
-}
-- LOCN 50 -50 true
-- STYLE abstract
--
import SS
--
-- LOCN 800 50 true
-- STYLE 
{-
DNA and the Genetic Code
-}
-- LOCN 50 75 true
-- STYLE title
--
picFile "misc/dna.png"
--
-- LOCN 50 250 true
-- STYLE 
{-
Acknowledgment:
This diagram, originally by Martin Chaplin, is 
taken from:
    http://en.wikipedia.org/wiki/Base_pair 
(part of the Wikipedia online encyclopedia).
-}
-- LOCN 675 425 true
-- STYLE comment
{-
A molecule of DNA is made up of a sequence of "base-pairs" each consisting of
a pair of bases (abbreviated as AT, TA, CG or GC) suspended between backbones
formed of sugar and phosphate groups, like so:
-}
-- LOCN 50 150 true
-- STYLE comment
{-
Representation
-}
-- LOCN 75 775 true
-- STYLE heading1
{-
The library module DNA_lib defines a collection of types and functions for representing
and manipulating molecules of DNA:
-}
-- LOCN 75 825 true
-- STYLE comment
--
import DNA_lib
--
-- LOCN 75 875 true
-- STYLE 
{-
The genetic content of a molecule of DNA can be described as a sequence of bases
(read in 5' to 3' direction on the "ascending" strand). For example:
-}
-- LOCN 75 950 true
-- STYLE comment
--
s = "ATCGATGACTGGGTAGGG"
--
-- LOCN 75 1000 true
-- STYLE hide
{-
Such a string can be parsed using the function
        parseDNA :: String -> DNA
to yield a representation (of type DNA) of the molecule
-}
-- LOCN 75 1075 true
-- STYLE comment
--
dna = parseDNA s
--
-- LOCN 75 1150 true
-- STYLE hide
{-
The  showDNA function shows a value of type DNA in string form:
-}
-- LOCN 75 1200 true
-- STYLE comment
--
pic = displayDNA dna
--
-- LOCN 75 1375 true
-- STYLE 
--
showDNA dna
--
-- LOCN 75 1225 true
-- STYLE 
{-
. . . and the displayDNA and displayDNA' functions shows it pictorially:
-}
-- LOCN 75 1325 true
-- STYLE comment
--
displayDNA' dna
--
-- LOCN 375 1375 true
-- STYLE 
--
scale 0.5 pic
--
-- LOCN 675 1375 true
-- STYLE 
{-
The Genetic Code
-}
-- LOCN 50 2000 true
-- STYLE heading1
{-
The sequence of bases in a molecule of DNA give rise to a sequence of amino acids by a process
known as transcription.
-}
-- LOCN 50 2050 true
-- STYLE comment
{-
There are some 20 different amino acids that can be synthesised; these are represented by
the datatype
       data Amino = Asp | . . . | Trp
-}
-- LOCN 50 2125 true
-- STYLE comment
{-
The function
        showAmino :: Amino -> String
maps these to their names. For example:
-}
-- LOCN 50 2225 true
-- STYLE comment
--
showAmino Asp
--
-- LOCN 125 2300 true
-- STYLE 
--
showAmino Cys
--
-- LOCN 325 2300 true
-- STYLE 
--
showAmino Trp
--
-- LOCN 700 2300 true
-- STYLE 
{-
The sequence of bases, taken in groups of three (called "codons"), encode the individual
amino acids:
        data Base = A | C | G | T

        type Codon = (Base, Base, Base)

        encode :: Codon -> Maybe Amino
-}
-- LOCN 50 2400 true
-- STYLE comment
{-
Not all codons code for amino acids; some define START and STOP codes (here represented by Nothing
values) that control the transcription process. For example:
-}
-- LOCN 50 2575 true
-- STYLE comment
--
encode (C, T, A)
--
-- LOCN 125 2625 true
-- STYLE 
--
encode (A, G, T)
--
-- LOCN 350 2625 true
-- STYLE 
--
encode (T, A, G)
--
-- LOCN 550 2625 true
-- STYLE 
{-
The outcome of the transcription process can be described by:

        type Junk = [Base]                -- some of the DNA does not code for amino acids

        type Gene = [Amino]               -- a sequence of amino acids is called a gene

        data Transcript = Transcript [Either Junk Gene]
                                          -- overall, a strand of DNA defines a sequence of interleaved
                                          -- subsequences of junk and genes
-}
-- LOCN 50 2725 true
-- STYLE commentM
{-
The process of transcription is defined by the function
        transcribe :: DNA -> Transcript
-}
-- LOCN 50 2950 true
-- STYLE comment
{-
For example:
-}
-- LOCN 50 3025 true
-- STYLE comment
--
transcript = transcribe dna
--
-- LOCN 50 3075 true
-- STYLE 
{-
The function
        display :: Transcript -> Pic
displays a transcript. For example:
-}
-- LOCN 50 3175 true
-- STYLE comment
--
displayTranscript transcript
--
-- LOCN 50 3275 true
-- STYLE 
{-
Some junk
-}
-- LOCN 250 3350 true
-- STYLE comment
{-
The start of the encoding (trigged by the sequence A T G)
-}
-- LOCN 250 3450 true
-- STYLE comment
{-
A couple of amino acids
-}
-- LOCN 250 3600 true
-- STYLE comment
{-
The end of the encoding (trigged by the codon T A G)
-}
-- LOCN 250 3725 true
-- STYLE comment
{-
Some junk
-}
-- LOCN 250 3800 true
-- STYLE comment
{-
The function
        displayAnnotDNA :: DNA -> Pic
displays a molecule of DNA with its transcript alongside it. For example:
-}
-- LOCN 50 3950 true
-- STYLE comment
--
displayAnnotDNA  dna
--
-- LOCN 50 4025 true
-- STYLE 
--
dna
--
-- LOCN 50 4850 true
-- STYLE 
{-
Direct manipulation of DNA
-}
-- LOCN 50 4675 true
-- STYLE heading1
{-
instance Edit DNA where {

  style dna = View displayAnnotDNA Picture;

  edit dna path = [("Delete base",  deleteFn),
                   ("Insert a gap", insertFn),
                   ("Set to A",     setToFn "A"),
                   ("Set to C",     setToFn "C"),
                   ("Set to G",     setToFn "G"),
                   ("Set to T",     setToFn "T"),
                   ("Unset",        setToFn "X"),
                   ("Evaluate",     evaluateFn)]
}
-}
-- LOCN 700 4725 true
-- STYLE commentM
{-
By defining an instance (see side box) of the Edit type class for the type DNA, the
In the DNAlib module, an instance of the Edit type class is defined for the type DNA
(along the lines shown in the side box). By this means:
    - a value of type DNA is, by default displayed pictorially, and
    - the user can click on any base pair in the diagram and change it.
-}
-- LOCN 50 4725 true
-- STYLE comment
{-
Try double-clicking (to select) one of the base pairs in this region
of the molecule, then right-clicking to drop-down a menu of possible
operations.
-}
-- LOCN 425 5125 true
-- STYLE warning2
{-
Links to related modules
-}
-- LOCN 50 5525 true
-- STYLE heading2
{-
This module presents a more realistically sized DNA molecule
on which the user can experiment with SNPs.
-}
-- LOCN 50 5575 true
-- STYLE open1 "DNA_SNIP"
{-
This module is a tidied-up version of the original module
that was interactively developed and from which the DNA_lib
library module was subsequently abstracted. It provides
a good illustration of the incremental, "document and test
as you go" approach that this document-centered mode
of programming facilitates.
-}
-- LOCN 50 5650 true
-- STYLE open1 "DNA_DEVELOP"
{-
Back to previous document
-}
-- LOCN 50 5800 true
-- STYLE open1 ""
{-
Back to previous document
-}
-- LOCN 800 100 true
-- STYLE open ""
{-
SNP experimentation
-}
-- LOCN 800 150 true
-- STYLE open "DNA_SNIP"
{-
Interactive development of
the DNAlib module.
-}
-- LOCN 800 200 true
-- STYLE open "DNA_DEVELOP"
